from typing import Any, Dict
import requests
from abc import abstractmethod
from ..wrapper_interfaces.model_wrapper import ModelWrapper

import logging


class WrapperException(Exception):
    pass


class RestAPIWrapper(ModelWrapper):
    def __init__(self, name: str, logger: logging.Logger = logging.getLogger()) -> None:
        self.logger = logger
        self.name = name

    def _get_headers(self) -> Dict[str, str]:
        return {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self._get_api_key()}"
        }

    @abstractmethod
    def _get_api_key(self) -> str:
        pass

    @abstractmethod
    def _get_body(self, prompt_message: str) -> Any:
        pass

    @abstractmethod
    def _get_api_url(self) -> str:
        pass

    def get_name(self) -> str:
        return self.name

    def prompt(self, prompt_message: str) -> Any:
        """
        raises GPTWrapperException if request failed
        returns string response from chatgpt completions api
        """
        self.logger.debug(f'Sending prompt to API: {prompt_message}')
        response = requests.post(
            self._get_api_url(),
            headers=self._get_headers(),
            data=self._get_body(prompt_message)
        )

        try:
            response.raise_for_status()
        except Exception as e:
            error_message = f"Request to api endpoint: {self._get_api_url()} failed.\n" \
                            f"Response message: {response.text}.\n" \
                            f"Exception caught: {e}"
            self.logger.error(error_message)
            raise WrapperException(error_message)

        return response.json()



from pydantic import BaseModel


class AIModel(BaseModel):
    id: int
    name: str
    type: str
    description: str

class TextualModelPromptRequest(BaseModel):
    model_name: str
    prompt: str

class TextualModelPromptResponse(BaseModel):
    response: str


import json
import os
from abc import ABC, abstractmethod
from typing import Dict, Type, List, Union, Any

FILE_NAME = ".saved_data.json"
MODULE_DIR = os.path.dirname(os.path.abspath(__file__))
FILE_PATH = os.path.join(MODULE_DIR, FILE_NAME)


class JsonConvertable(ABC):
    @abstractmethod
    def to_json(self) -> Dict[str, Union[str, int]]:
        pass

    @classmethod
    @abstractmethod
    def create_from_json(cls, json_key: str, json_data: Dict[str, str]) -> Any:
        pass


class JsonDataManager:
    @staticmethod
    def save(nested_path: List[str], item: JsonConvertable, file_path: str = FILE_PATH) -> None:
        json_value = item.to_json()

        data = {}
        if os.path.isfile(file_path):
            with open(file_path, 'r') as f:
                try:
                    data = json.load(f)
                except json.JSONDecodeError:
                    pass

        current_data = data
        for key in nested_path[:-1]:
            current_data = current_data.setdefault(key, {})

        current_data[nested_path[-1]] = json_value

        with open(file_path, 'w') as f:
            json.dump(data, f, indent=2)

    @staticmethod
    def load(nested_path: List[str], cls: Type[JsonConvertable], file_path: str = FILE_PATH) -> Any:
        data = JsonDataManager.get_json_value(nested_path, file_path)
        if len(nested_path) == 0:
            json_key = ""
        else:
            json_key = nested_path[-1]
        return cls.create_from_json(json_key, data)

    @staticmethod
    def get_json_value(nested_path: List[str], file_path: str = FILE_PATH) -> Any:
        if not os.path.exists(file_path):
            raise FileNotFoundError("Saved data file cannot be found.")

        with open(file_path, "r") as f:
            data = json.load(f)
            for key in nested_path:
                try:
                    data = data[key]
                except KeyError:
                    raise KeyError(f"Key '{key}' not found in JSON data")
            return data



from abc import ABC
from typing import List


class ModelIO(ABC):
    pass


class TextualIO(ModelIO):
    def __init__(self, text: str):
        self.text = text

    def get_text(self) -> str:
        return self.text

    def __repr__(self):  # type: ignore
        return self.get_text()


class FlowTextIO(TextualIO):
    def __init__(self, text: str):
        super().__init__(text)
        self.pretty_text = text

    def __repr__(self) -> str:
        return self.pretty_text

    def __iadd__(self, other):  # type: ignore
        if isinstance(other, TextualIO):
            self.text += '\n' + other.get_text()
            self.pretty_text += '\n\n' + ('*' * 100) + '\n\n' + other.get_text()
            return self

        else:
            raise TypeError(f"unsupported operand type(s) for +: '{type(self).__name__}' and '{type(other).__name__}'")


class ImageIO(ModelIO):
    def __init__(self, paths: List[str]):
        self.paths = paths

    def image_paths(self):  # type: ignore
        return self.paths

    def __repr__(self):  # type: ignore
        return f"image paths: {self.paths}"

class FileIO(ModelIO):
    pass



from typing import Dict, Any, Union
import json

from .rest_api_wrapper import RestAPIWrapper
import logging

from ...model_entities.models_io.base_model_io import ModelIO, TextualIO

COMPLETIONS_API_URL = "https://api.openai.com/v1/chat/completions"


class GPTWrapper(RestAPIWrapper):
    def __init__(self, name: str, api_key: str, model: str, logger: logging.Logger = logging.getLogger()) -> None:
        if not api_key:
            raise ValueError("api key cannot be None.")
        if not model:
            raise ValueError("model cannot be None.")
        super().__init__(name, logger)
        self.api_url = COMPLETIONS_API_URL
        self.api_key = api_key
        self.model = model

    def to_json(self) -> Dict[str, Union[str, int]]:
        return {
            'api key': self.api_key,
            'model': self.model,
            '_class_type': self.get_class_type_field()
        }

    @classmethod
    def create_from_json(cls, json_key: str, json_data: Dict[str, str]) -> Any:
        return GPTWrapper(json_key, json_data['api key'], json_data['model'])

    @classmethod
    def get_class_type_field(cls) -> str:
        return "chat_gpt"

    def _get_api_url(self) -> str:
        return self.api_url

    def _parse_response(self, response: Dict[str, Any]) -> Any:
        return {'response': response['choices'][0]['message']['content']}

    def _get_api_key(self) -> str:
        return self.api_key

    def _get_body(self, prompt_message: str) -> Any:
        return json.dumps({
            "model": self.model,
            "messages": [
                {"role": "user",
                 "content": prompt_message}
            ]
        })

    def process(self, model_input: ModelIO) -> ModelIO:
        if not isinstance(model_input, TextualIO):
            raise TypeError(f"expected type {type(TextualIO)} but got {type(model_input)}")

        response = self.prompt(model_input.text)
        chat_answer = response['choices'][0]['message']['content']
        return TextualIO(chat_answer)



from fastapi import FastAPI
from omnibridge_api.app.core.config import settings
from omnibridge_api.app.api_v1.api import api_router


app = FastAPI(
    title=settings.PROJECT_NAME, openapi_url=f"{settings.API_V1_STR}/openapi.json"
)


app.include_router(api_router, prefix=settings.API_V1_STR)



import argparse

from omnibridge.flows.branching_flow import BranchingFlow
from omnibridge.flows.flow_loader import FlowLoader
from omnibridge.flows.sequential_flow import SequentialFlow
from omnibridge.model_entities.models_io.base_model_io import TextualIO
from omnibridge.saved_data.json_data_manager import JsonDataManager
from omnibridge.wrappers.api_key import ApiKey
from omnibridge.cli.banner import banner
from omnibridge.wrappers.wrapper_instances.dalle_wrapper import DALLEWrapper
from omnibridge.wrappers.wrapper_instances.gpt_wrapper import GPTWrapper
from omnibridge.wrappers.wrapper_instances.type_name_to_wrapper import ModelLoader


def run() -> int:
    parser = argparse.ArgumentParser(description='AI integration tool.')
    parser.add_argument('-p', '--prompt', help="prompt for model", type=str, default=[])
    parser.add_argument('-l', "--load-config", help="absolute path to models configuration file")
    parser.add_argument('-n', "--model-name", help="which model to use.")

    subparsers = parser.add_subparsers(help='sub-command help', dest='command')
    api_key_parser = subparsers.add_parser("add-key", help="Add an api key.")
    api_key_parser.add_argument('-n', '--name', help="name of the api key.", type=str, required=True)
    api_key_parser.add_argument('-v', '--value', help="value of the api key.", type=str, required=True)

    add_chatgpt_model_parser = subparsers.add_parser("add-chatgpt", help="add chatgpt model connection details.")
    add_chatgpt_model_parser.add_argument('-n', '--name', type=str, required=True,
                                          help="name of the model, e.g. my_gpt4.")
    add_chatgpt_model_parser.add_argument('-k', '--key', type=str, required=True, help="api key name.")
    add_chatgpt_model_parser.add_argument('-m', '--model', type=str, default="gpt-3.5-turbo",
                                          help="model, e.g. 3.5 or 4.")

    add_chatgpt_model_parser = subparsers.add_parser("add-dalle", help="add dalle model.")
    add_chatgpt_model_parser.add_argument('-n', '--name', type=str, required=True, help="name of the model.")
    add_chatgpt_model_parser.add_argument('-k', '--key', type=str, required=True, help="api key name.")
    add_chatgpt_model_parser.add_argument('--num-images', type=str, default="4",
                                          help="number of images per prompt, default 4.")
    add_chatgpt_model_parser.add_argument('-r', '--res', type=str, default="256x256", help="resolution of images.")

    add_chatgpt_model_parser = subparsers.add_parser("add-flow", help="add flow.")
    add_chatgpt_model_parser.add_argument('-n', '--name', type=str, required=True, help="name of the flow.")
    add_chatgpt_model_parser.add_argument('-t', '--type', type=str, choices=['seq', 'branching'], default='branching',
                                          help="type of the flow.")
    add_chatgpt_model_parser.add_argument('-m', '--model', type=str,
                                          help="name of the model to use.")
    add_chatgpt_model_parser.add_argument('-i', '--instruction', nargs='*', type=str,
                                          help="instruct the model to do something with the input.")
    add_chatgpt_model_parser.add_argument('--multi', nargs='*', type=str,
                                          help="use multiple models. Currently only supported in sequential flow.")

    add_chatgpt_model_parser = subparsers.add_parser("run-flow", help="run flow.")
    add_chatgpt_model_parser.add_argument('-n', '--name', type=str, required=True, help="name of the flow.")
    add_chatgpt_model_parser.add_argument('-p', '--prompt', help="prompt for flow", type=str, required=True)

    args = vars(parser.parse_args())
    if args['command'] == 'run-flow':
        flow = FlowLoader.load_flow(args['name'])
        flow_input = TextualIO(args['prompt'])
        flow_output = flow.process(flow_input)
        print(flow_output)
        return 0
    if args['command'] == 'add-flow':
        if args['type'] == 'branching':
            instructions = args['instruction']
            model = ModelLoader.load_model(args['model'])
            b_flow = BranchingFlow(args['name'], model, instructions)
            JsonDataManager.save(["flows", args["name"]], b_flow)
            return 0
        if args['type'] == 'seq':
            models = [ModelLoader.load_model(model_name) for model_name in args['multi']]
            flow = SequentialFlow(args['name'], models)
            JsonDataManager.save(["flows", args["name"]], flow)
            return 0
    if args['command'] == 'add-key':
        api_key = ApiKey(args['value'])
        JsonDataManager.save(["api keys", args['name']], api_key)
        return 0
    elif args['command'] == 'add-chatgpt':
        api_key: ApiKey = JsonDataManager.load(["api keys", args['key']], ApiKey)
        wrapper: GPTWrapper = GPTWrapper(args['name'], api_key.value, args['model'])
        JsonDataManager.save(["models", args['name']], wrapper)
        return 0
    elif args['command'] == 'add-dalle':
        api_key: ApiKey = JsonDataManager.load(["api keys", args['key']], ApiKey)
        wrapper: DALLEWrapper = DALLEWrapper(name=args['name'], api_key=api_key.value,
                                             number_of_images=args['num_images'], resolution=args['res'])
        JsonDataManager.save(["models", args['name']], wrapper)
        return 0
    elif args['prompt'] and args['model_name']:
        wrapper = ModelLoader.load_model(model_name=args['model_name'])
        response = wrapper.process(TextualIO(args['prompt']))
        print(response)
        return 0

    return 1


def main():
    print(banner)
    exit(run())


if __name__ == '__main__':
    main()



from typing import List, Dict

from omnibridge.flows.flow import Flow
from omnibridge.model_entities.models_io.base_model_io import ModelIO
from omnibridge.wrappers.wrapper_instances.type_name_to_wrapper import ModelLoader
from omnibridge.wrappers.wrapper_interfaces.model_wrapper import ModelWrapper


class SequentialFlow(ModelWrapper, Flow):
    def __init__(self, name: str, models: List[ModelWrapper] = None):
        self.name = name
        self.models = models
        if self.models is None:
            self.models = []

    def add(self, new_model: ModelWrapper, instruction: str) -> None:
        self.models.append(new_model)

    def process(self, model_input: ModelIO) -> ModelIO:
        if len(self.models) == 0:
            raise RuntimeError("Cannot run empty flow.")

        model_input = model_input
        for model in self.models:
            model_output = model.process(model_input)
            print(model_output)
            model_input = model_output

        return model_output

    def to_json(self) -> Dict[str, str]:
        return {
            '_class_type': self.get_class_type_field(),
            'models': ', '.join(model.get_name() for model in self.models)
        }

    @classmethod
    def create_from_json(cls, json_key, json_data: Dict[str, str]):
        name = json_key
        models_names = json_data['models'].split(', ')
        models = [ModelLoader.load_model(model_name) for model_name in models_names]
        return SequentialFlow(name, models)

    def get_name(self) -> str:
        return self.name

    @classmethod
    def get_class_type_field(cls):
        return "sequential"
